//
//  ImageCacheItUtils.swift
//  NativeScript ImageCacheIt
//
//  Created by Osei Fortune on 19/10/2020.
//

import Foundation
import UIKit

@objcMembers
@objc(ImageCacheItUtils)
public class ImageCacheItUtils: NSObject {
    private static var queueCount: Int64 = 0
    public static func createConcurrentQueue(_ name: String) -> DispatchQueue {
        return DispatchQueue(label: name, attributes: .concurrent)
    }

    public static func createImageOverlay(_ image: UIImage?, _ color: String) -> UIImage? {

        let rgba = color.replacingOccurrences(of: "rgba(", with: "").replacingOccurrences(of: ")", with: "").replacingOccurrences(of: " ", with: "").split(separator: ",")
        let formatter = NumberFormatter()
        let red = CGFloat(Int32(truncating: formatter.number(from: String(rgba[0])) ?? 255)) / 255.0
        let green = CGFloat(Int32(truncating:formatter.number(from: String(rgba[1])) ?? 255)) / 255.0
        let blue = CGFloat(Int32(truncating:formatter.number(from: String(rgba[2])) ?? 255)) / 255.0
        let alpha = CGFloat(truncating: formatter.number(from: String(rgba[3])) ?? 1)

        return createImageOverlay(image, image?.size.width ?? 0 , image?.size.height ?? 0, red, green, blue, alpha)
    }

    public static func createImageOverlay(_ image: UIImage?,_ width: CGFloat,_ height: CGFloat,_ red: CGFloat,_ green: CGFloat,_ blue: CGFloat,_ alpha: CGFloat) -> UIImage? {
        if(image == nil){
            return image
        }
        let zero = CGFloat(0)
        let rect = CGRect(x: zero, y: zero, width: CGFloat(width), height: CGFloat(height))
        let width = image!.cgImage!.width
        let height = image!.cgImage!.height

        let colourSpace = CGColorSpaceCreateDeviceRGB()
        let imageContext =  CGContext(data: nil,
                                      width: width,
                                      height: height,
                                      bitsPerComponent: 8,
                                      bytesPerRow: width * 4,
                                      space: colourSpace,
                                      bitmapInfo: CGImageAlphaInfo.premultipliedFirst.rawValue | CGBitmapInfo.byteOrder32Little.rawValue)

        imageContext?.draw(image!.cgImage!, in: rect)
        let maskImage = image!.cgImage

        if(maskImage != nil){
            imageContext?.clip(to: rect, mask: maskImage!)
        }
        imageContext?.setFillColor(red: red, green: green, blue: blue, alpha: alpha)
        imageContext?.fill(rect)
        let outputImage = imageContext!.makeImage()
        return UIImage(cgImage: outputImage!)
    }

    private static var filterCache: [String: CIFilter] = [:]

    private static func getFilterByName(_ name: String,_ image: UIImage?) -> CIFilter?{
        let filter: CIFilter? = CIFilter(name:name)
        if (filter == nil){
            return nil
        }

        if let ref = image {
            if (ref.ciImage != nil) {
                filter?.setValue(ref.ciImage!, forKey: kCIInputImageKey)
                filter?.setValue(nil, forKey: CIImageOption.colorSpace.rawValue)
            } else {
                if (ref.cgImage != nil) {
                    filter?.setValue(CIImage(cgImage: ref.cgImage!), forKey: kCIInputImageKey)
                }
            }
        }
        return filter;
    }

    private static func createAlphaImage(_ image: UIImage?, _ alpha: CGFloat) -> UIImage?{
        if(image == nil){
            return image
        }
        let zero = CGFloat(0)
        let width = image!.cgImage!.width
        let height = image!.cgImage!.height
        let rect = CGRect(x: zero, y: zero, width: CGFloat(width), height: CGFloat(height))
        let colourSpace = CGColorSpaceCreateDeviceRGB()
        let imageContext =  CGContext(data: nil,
                                      width: width,
                                      height: height,
                                      bitsPerComponent: 8,
                                      bytesPerRow: width * 4,
                                      space: colourSpace,
                                      bitmapInfo: CGImageAlphaInfo.premultipliedFirst.rawValue | CGBitmapInfo.byteOrder32Little.rawValue)
        imageContext?.setAlpha(alpha)
        imageContext?.draw(image!.cgImage!, in: rect)
        guard let outputImage = imageContext?.makeImage() else {return nil}
        return UIImage(cgImage: outputImage)
    }

    public static func applyProcessing(_ ctx: CIContext,_ image: UIImage?, _ filters: [String: Any], _ callback: @escaping (UIImage?)-> Void) {
        guard image != nil else {
            DispatchQueue.main.async {
                callback(nil)
            }
            return
        }
        let queue = createConcurrentQueue("ImageCacheUtils-Queue-" + String(queueCount))
        queue.async {
            let filter = filters["filter"] as? String
            let filteredImage = applyFilters(image, ctx,filter)
            if let overLayer = filters["overlayColor"] as? String {
                let overlayImage = createImageOverlay(filteredImage, overLayer)
                DispatchQueue.main.async {
                    callback(overlayImage)
                }
            }else {
                DispatchQueue.main.async {
                    callback(filteredImage)
                }
            }
        }
    }

    private static func applyFilters(_ image: UIImage?, _ ctx: CIContext,_ filter: String?) -> UIImage? {
        if(image == nil){
            return nil
        }
        var realImage = image!
        if(filter != nil){
            func getValue(_ value: String) -> String? {
                guard let startIndex = value.firstIndex(of: "(") else {return nil}
                let start = value.index(after: startIndex)
                guard let end = value.firstIndex(of: ")") else {return nil}
                return String(value[ start ... end]);
            }
            func createFilterWithName(_ value: String) -> CIFilter?{
                return getFilterByName(value, image)
            }

            let imageFilters = filter?.split(separator: " ").map {
                String($0)
            } ?? []
            for imageFilter in imageFilters {
                let value = getValue(imageFilter)

                switch imageFilter {
                case "blur":
                    var width: CGFloat?
                    if let size = value {
                        if(size.contains("%")){
                            if let percent = Float(size.replacingOccurrences(of: "%", with: "")) {
                                width = CGFloat((CGFloat(percent) / 100.0) * realImage.size.width)
                            }
                        }else if(size.contains("px")){
                            if let px = Float(size.replacingOccurrences(of: "px", with: "")) {
                                width = CGFloat(px) * UIScreen.main.scale
                            }
                        }else if(size.contains("dip")) {
                            if let dip = Float(size.replacingOccurrences(of: "dip", with: "")) {
                                width = CGFloat(dip)
                            }

                        }else{
                            if let raw = Float(size){
                                width = CGFloat(raw)
                            }
                        }

                        if var realWidth = width {
                            if (realWidth > 25) {
                                realWidth = 25
                            }

                            if (realWidth > -1) {
                                let blurFilter = createFilterWithName("CIGaussianBlur")
                                blurFilter?.setValue(width, forKey: kCIInputRadiusKey)
                                if let blurredImg = blurFilter?.value(forKey: kCIOutputImageKey) as? CIImage {
                                    if let cgImage = ctx.createCGImage(blurredImg, from: blurredImg.extent) {
                                        realImage = UIImage(cgImage: cgImage)
                                    }

                                }
                            }
                        }
                    }
                    break
                case "contrast":
                    if let size = value {
                        if (size.contains("%")) {
                            if let contrast = Float(size.replacingOccurrences(of: "%", with: "")) {
                                _ = contrast / 100.0
                                let contrastFilter = createFilterWithName("CIColorControls");
                                contrastFilter?.setValue(contrast, forKey: kCIInputContrastKey)
                                if let contrastImg: CIImage = contrastFilter?.value(forKey: kCIOutputImageKey)  as? CIImage{
                                    if let cgImage = ctx.createCGImage(contrastImg, from: contrastImg.extent) {
                                        realImage = UIImage(cgImage: cgImage)
                                    }
                                }

                            }
                        }
                    }
                    break
                case "brightness":
                    if let size = value {
                        if (size.contains("%")) {
                            if let brightness = Float(size.replacingOccurrences(of: "%", with: "")) {
                                /* if (brightness >= 0 && brightness < 1) {
                                 brightness = -1 + brightness;
                                 }*/

                                let brightnessFilter = createFilterWithName("CIColorControls");
                                brightnessFilter?.setValue(brightness, forKey: kCIInputBrightnessKey)
                                if let contrastImg = brightnessFilter?.value(forKey: kCIOutputImageKey) as? CIImage {
                                    if let cgImage = ctx.createCGImage(contrastImg, from: contrastImg.extent) {
                                        realImage = UIImage(cgImage: cgImage)
                                    }
                                }

                            }
                        }
                    }
                    break
                case "greyscale" , "grayscale":
                    var grayscale: CGFloat = 0
                    if let size = value {
                        if (size.contains("%")) {
                            if let realSize = Float(size.replacingOccurrences(of: "%", with: "")) {
                                grayscale = CGFloat(realSize / 100.0)
                            }
                        } else if (size.contains(".")) {
                            if let realSize = Float(size) {
                                grayscale = CGFloat(realSize)
                            }
                        } else {
                            if let realSize = Float(size) {
                                grayscale = CGFloat(realSize)
                            }
                        }

                        if (grayscale > 1) {
                            grayscale = 1;
                        }

                        grayscale = 1 - grayscale;

                        let grayscaleFilter = createFilterWithName("CIColorControls")
                        grayscaleFilter?.setValue(grayscale, forKey: kCIInputSaturationKey)
                        if let grayscaleImg = grayscaleFilter?.value(forKey: kCIOutputImageKey) as? CIImage {
                            if let cgImage = ctx.createCGImage(grayscaleImg, from: grayscaleImg.extent){
                                realImage = UIImage(cgImage: cgImage)
                            }
                        }
                    }
                    break
                case "invert":
                    // TODO handle value
                    let invertFilter = createFilterWithName("CIColorInvert")
                    if let invertImg = invertFilter?.value(forKey: kCIOutputImageKey) as? CIImage{
                        if let cgImage = ctx.createCGImage(invertImg,from: invertImg.extent) {
                            realImage = UIImage(cgImage:cgImage)
                        }
                    }
                    break
                case "sepia":
                    if let size = value {
                        if let sepia = Float(size.replacingOccurrences(of: "%", with: "")) {
                            let realSize = sepia / 100
                            let sepiaFilter = createFilterWithName("CISepiaTone")
                            sepiaFilter?.setValue(realSize, forKey: kCIInputIntensityKey)
                            if let sepiaImg = sepiaFilter?.value(forKey: kCIOutputImageKey) as? CIImage {
                                if let cgImage = ctx.createCGImage(sepiaImg,from: sepiaImg.extent) {
                                    realImage = UIImage(cgImage: cgImage)
                                }
                            }
                        }
                    }
                    break
                case "opacity":
                    if let size = value {
                        var alpha: CGFloat = 1
                        if (size.contains("%")) {
                            if let realSize = Float(size.replacingOccurrences(of: "%", with: "")) {
                                alpha = CGFloat(realSize / 100)
                            }
                        } else if (size.contains(".")) {
                            if let realSize = Float(size) {
                                alpha = CGFloat(realSize)
                            }
                        } else {
                            if let realSize = Float(size) {
                                alpha = CGFloat(realSize)
                            }
                        }
                        if let newImage = createAlphaImage(realImage, alpha) {
                            realImage = newImage
                        }
                    }
                    break
                case "hue":
                    let hueFilter = createFilterWithName("CIHueAdjust")
                    var hue = 0
                    if let realValue = value {
                        if (realValue.contains("deg")) {
                            if let realHue = Int(realValue.replacingOccurrences(of: "deg", with: "")){
                                hue = realHue
                            }
                        } else if (realValue.contains("turn")) {
                            if let realHue = Int(realValue.replacingOccurrences(of: "turn", with: "")){
                                hue = realHue * 360
                            }
                        }
                        hueFilter?.setValue(hue, forKey: kCIInputAngleKey)

                        if let hueImg = hueFilter?.value(forKey: kCIOutputImageKey) as? CIImage {
                            if let cgImage = ctx.createCGImage(hueImg,from: hueImg.extent){
                                realImage = UIImage(cgImage: cgImage)
                            }
                        }
                    }
                    break
                case "saturate":
                    let saturateFilter = createFilterWithName("CIColorControls")
                    var saturate: CGFloat?
                    if let size = value {
                        if(size.contains("%")){
                            if let realSize = Float(size.replacingOccurrences(of: "%", with: "")){
                                saturate = CGFloat(realSize / 100)
                            }
                        }else if(size.contains(".")){
                            if let realSize = Float(size){
                                saturate = CGFloat(realSize)
                            }
                        }else {
                            if let realSize = Float(size){
                                saturate = CGFloat(realSize)
                            }
                        }
                    }

                    if let sat = saturate {
                        saturateFilter?.setValue(sat, forKey: kCIInputSaturationKey)
                        if let saturateImg = saturateFilter?.value(forKey: kCIOutputImageKey) as? CIImage {
                            if let cgImage = ctx.createCGImage(saturateImg, from: saturateImg.extent){
                                realImage = UIImage(cgImage: cgImage)
                            }
                        }
                    }
                    break
                default:
                    // NOOP
                    break
                }
            }
        }
        return realImage
    }
}

